@@ -1,7 +1,7 @@
 /*
 	libmpg123: MPEG Audio Decoder library
 
-	copyright 1995-2009 by the mpg123 project - free software under the terms of the LGPL 2.1
+	copyright 1995-2010 by the mpg123 project - free software under the terms of the LGPL 2.1
 	see COPYING and AUTHORS files in distribution or http://mpg123.org
 
 */
@@ -66,6 +66,23 @@ static void frame_buffercheck(mpg123_handle *fr)
 	/* When we have no accurate position, gapless code does not make sense. */
 	if(!fr->accurate) return;
 
+	/* Important: We first cut samples from the end, then cut from beginning (including left-shift of the buffer).
+	   This order works also for the case where firstframe == lastframe. */
+
+	/* The last interesting (planned) frame: Only use some leading samples.
+	   Note a difference from the below: The last frame and offset are unchanges by seeks.
+	   The lastoff keeps being valid. */
+	if(fr->lastframe > -1 && fr->num >= fr->lastframe)
+	{
+		/* There can be more than one frame of padding at the end, so we ignore the whole frame if we are beyond lastframe. */
+		off_t byteoff = (fr->num == fr->lastframe) ? samples_to_bytes(fr, fr->lastoff) : 0;
+		if((off_t)fr->buffer.fill > byteoff)
+		{
+			fr->buffer.fill = byteoff;
+		}
+		debug1("Cut frame buffer on end of stream, fill now %"SIZE_P" bytes.", (size_p)fr->buffer.fill);
+	}
+
 	/* The first interesting frame: Skip some leading samples. */
 	if(fr->firstoff && fr->num == fr->firstframe)
 	{
@@ -86,17 +103,6 @@ static void frame_buffercheck(mpg123_handle *fr)
 		   So it is safe to null it here (and it makes the if() decision abort earlier). */
 		fr->firstoff = 0;
 	}
-	/* The last interesting (planned) frame: Only use some leading samples.
-	   Note a difference from the above: The last frame and offset are unchanges by seeks.
-	   The lastoff keeps being valid. */
-	if(fr->lastoff && fr->num == fr->lastframe)
-	{
-		off_t byteoff = samples_to_bytes(fr, fr->lastoff);
-		if((off_t)fr->buffer.fill > byteoff)
-		{
-			fr->buffer.fill = byteoff;
-		}
-	}
 }
 #endif
 
@@ -916,7 +922,14 @@ int attribute_align_arg mpg123_feed(mpg123_handle *mh, const unsigned char *in,
 		if(in != NULL)
 		{
 			if(feed_more(mh, in, size) != 0) return MPG123_ERR;
-			else return MPG123_OK;
+			else
+			{
+				/* The need for more data might have triggered an error.
+				   This one is outdated now with the new data. */
+				if(mh->err == MPG123_ERR_READER) mh->err = MPG123_OK;
+
+				return MPG123_OK;
+			}
 		}
 		else
 		{
@@ -1322,10 +1335,15 @@ off_t attribute_align_arg mpg123_length(mpg123_handle *mh)
 	else if(mh->rdat.filelen == 0) return mpg123_tell(mh); /* we could be in feeder mode */
 	else return MPG123_ERR; /* No length info there! */
 
+	debug1("mpg123_length: internal sample length: %"OFF_P, (off_p)length);
+
 	length = frame_ins2outs(mh, length);
+	debug1("mpg123_length: external sample length: %"OFF_P, (off_p)length);
 #ifdef GAPLESS
+	debug2("mpg123_length: begin_os = %"OFF_P", end_os = %"OFF_P, (off_p)mh->begin_os, (off_p)mh->end_os);
 	if(mh->end_os > 0 && length > mh->end_os) length = mh->end_os;
 	length -= mh->begin_os;
+	debug1("mpg123_length: after gapless correction: %"OFF_P, (off_p)length);
 #endif
 	return length;
 }
